import os
import time

import dask.diagnostics.progress
import odc.stac
import planetary_computer as pc
import pystac
import rioxarray  # seems unused but is needed
import xarray as xr
from shapely.geometry import shape
from tenacity import retry, stop_after_attempt, wait_random_exponential

from .cfg import BANDS_OF_INTEREST, COLLECTION_ID, MSPC_URL


@retry(wait=wait_random_exponential(max=3), stop=stop_after_attempt(2))
def get_item(id):
    if "/" not in id:
        uri = MSPC_URL + "/collections/" + COLLECTION_ID + "/items/" + id
    else:
        uri = id

    item = pystac.Item.from_file(uri)

    if uri.startswith(MSPC_URL):
        item = pc.sign(item)

    return item


def create_input(win_a, win_b, out, overwrite, bbox=None):
    """Main function for creating input for inference."""
    out = os.path.abspath(out)
    if os.path.exists(out) and not overwrite:
        print("Output file already exists, use -f to overwrite them. Exiting.")
        return

    # Ensure that the base directory exists
    os.makedirs(os.path.dirname(out), exist_ok=True)

    # Parse the bounding box
    if bbox is not None:
        bbox = list(map(float, bbox.split(",")))

    # Load the items
    identifiers = [win_a, win_b]
    items = []
    timestamp = None
    version = 0
    for i in identifiers:
        item = get_item(i)
        items.append(item)

        datetime = item.datetime or item.start_datetime or item.end_datetime
        if datetime and (not timestamp or datetime > timestamp):
            timestamp = datetime

        proc_version = item.properties.get(
            "processing:version", item.properties.get("s2:processing_baseline", 0)
        )
        try:
            proc_version = float(proc_version)
        except TypeError:
            proc_version = 0
        if proc_version > 0:
            if version > 0 and version != proc_version:
                print("Processing version of imagery differs. Exiting.")
                return
            version = proc_version

    shapes = [shape(item.geometry) for item in items]
    # Ensure the images intersect
    if not shapes[0].intersects(shapes[1]):
        print("The provided images do not intersect. Exiting.")
        return

    tic = time.time()
    data = odc.stac.load(
        [items[0], items[1]],
        bands=BANDS_OF_INTEREST,
        dtype="uint16",
        resampling="bilinear",
        bbox=bbox,
        chunks={"x": "auto", "y": "auto"},
    )

    data = (
        data.to_array(dim="band")
        .stack(bands=("time", "band"))
        .drop_vars("band")
        .transpose("bands", "y", "x")
    )

    if version < 3 or version >= 4:
        print(
            f"Processing version {version} unknown or untested (< 3.0 or >= 4.0). Inference quality might decrease."
        )
    if version >= 4:
        print(
            f"Rescaling data to processing version 3.0 from processing version {version}."
        )
        data = (data.astype("int32") - 1000).clip(min=0).astype("uint16")

    print("Writing output")
    with dask.diagnostics.progress.ProgressBar():
        data.rio.to_raster(
            out,
            driver="GTiff",
            compress="deflate",
            dtype="uint16",
            tiled=True,
            blockxsize=256,
            blockysize=256,
            tags={"TIFFTAG_DATETIME": timestamp.strftime("%Y:%m:%d %H:%M:%S")},
        )

    print(f"Finished merging and writing output in {time.time() - tic:0.2f} seconds")



import os
import time

import numpy as np
import torch
from lightning.pytorch.cli import LightningCLI
from torch.utils.data import DataLoader
from torchgeo.datamodules import BaseDataModule
from torchgeo.trainers import BaseTask
from torchmetrics import JaccardIndex, MetricCollection, Precision, Recall
from tqdm import tqdm

from ftw.datamodules import preprocess
from ftw.datasets import FTW
from ftw.metrics import get_object_level_metrics
from ftw.trainers import CustomSemanticSegmentationTask


def fit(config, ckpt_path, cli_args):
    """Command to fit the model."""
    print("Running fit command")

    # Construct the arguments for PyTorch Lightning CLI
    cli_args = ["fit", f"--config={config}"] + list(cli_args)

    # If a checkpoint path is provided, append it to the CLI arguments
    if ckpt_path:
        cli_args += [f"--ckpt_path={ckpt_path}"]

    print(f"CLI arguments: {cli_args}")

    # Best practices for Rasterio environment variables
    rasterio_best_practices = {
        "GDAL_DISABLE_READDIR_ON_OPEN": "EMPTY_DIR",
        "AWS_NO_SIGN_REQUEST": "YES",
        "GDAL_MAX_RAW_BLOCK_CACHE_SIZE": "200000000",
        "GDAL_SWATH_SIZE": "200000000",
        "VSI_CURL_CACHE_SIZE": "200000000",
    }
    os.environ.update(rasterio_best_practices)

    # Run the LightningCLI with the constructed arguments
    cli = LightningCLI(
        model_class=BaseTask,
        datamodule_class=BaseDataModule,
        seed_everything_default=0,
        subclass_mode_model=True,
        subclass_mode_data=True,
        save_config_kwargs={"overwrite": True},
        args=cli_args,  # Pass the constructed cli_args
    )

    print("Finished")


def test(
    model,
    dir,
    gpu,
    countries,
    postprocess,
    iou_threshold,
    out,
    model_predicts_3_classes,
    test_on_3_classes,
    temporal_options,
    cli_args,
):
    """Command to test the model."""
    print("Running test command")

    # Merge `test_model` function into this test command
    if torch.cuda.is_available():
        device = torch.device(f"cuda:{gpu}")
    else:
        device = torch.device("cpu")

    print("Loading model")
    tic = time.time()
    trainer = CustomSemanticSegmentationTask.load_from_checkpoint(
        model, map_location="cpu"
    )
    model = trainer.model.eval().to(device)
    print(f"Model loaded in {time.time() - tic:.2f}s")

    print("Creating dataloader")
    tic = time.time()

    ds = FTW(
        root=dir,
        countries=countries,
        split="test",
        transforms=preprocess,
        load_boundaries=test_on_3_classes,
        temporal_options=temporal_options,
    )
    dl = DataLoader(ds, batch_size=64, shuffle=False, num_workers=12)
    print(f"Created dataloader with {len(ds)} samples in {time.time() - tic:.2f}s")

    if test_on_3_classes:
        metrics = MetricCollection(
            [
                JaccardIndex(
                    task="multiclass", average="none", num_classes=3, ignore_index=3
                ),
                Precision(
                    task="multiclass", average="none", num_classes=3, ignore_index=3
                ),
                Recall(
                    task="multiclass", average="none", num_classes=3, ignore_index=3
                ),
            ]
        ).to(device)
    else:
        metrics = MetricCollection(
            [
                JaccardIndex(
                    task="multiclass", average="none", num_classes=2, ignore_index=3
                ),
                Precision(
                    task="multiclass", average="none", num_classes=2, ignore_index=3
                ),
                Recall(
                    task="multiclass", average="none", num_classes=2, ignore_index=3
                ),
            ]
        ).to(device)

    all_tps = 0
    all_fps = 0
    all_fns = 0
    for batch in tqdm(dl):
        images = batch["image"].to(device)
        masks = batch["mask"].to(device)
        with torch.inference_mode():
            outputs = model(images)

        outputs = outputs.argmax(dim=1)

        if model_predicts_3_classes:
            new_outputs = torch.zeros(
                outputs.shape[0], outputs.shape[1], outputs.shape[2], device=device
            )
            new_outputs[outputs == 2] = 0  # Boundary pixels
            new_outputs[outputs == 0] = 0  # Background pixels
            new_outputs[outputs == 1] = 1  # Crop pixels
            outputs = new_outputs
        else:
            if test_on_3_classes:
                raise ValueError(
                    "Cannot test on 3 classes when the model was trained on 2 classes"
                )

        metrics.update(outputs, masks)
        outputs = outputs.cpu().numpy().astype(np.uint8)
        masks = masks.cpu().numpy().astype(np.uint8)

        for i in range(len(outputs)):
            output = outputs[i]
            mask = masks[i]
            if postprocess:
                post_processed_output = out.copy()
                output = post_processed_output
            tps, fps, fns = get_object_level_metrics(
                mask, output, iou_threshold=iou_threshold
            )
            all_tps += tps
            all_fps += fps
            all_fns += fns

    results = metrics.compute()
    pixel_level_iou = results["MulticlassJaccardIndex"][1].item()
    pixel_level_precision = results["MulticlassPrecision"][1].item()
    pixel_level_recall = results["MulticlassRecall"][1].item()

    if all_tps + all_fps > 0:
        object_precision = all_tps / (all_tps + all_fps)
    else:
        object_precision = float("nan")

    if all_tps + all_fns > 0:
        object_recall = all_tps / (all_tps + all_fns)
    else:
        object_recall = float("nan")

    print(f"Pixel level IoU: {pixel_level_iou:.4f}")
    print(f"Pixel level precision: {pixel_level_precision:.4f}")
    print(f"Pixel level recall: {pixel_level_recall:.4f}")
    print(f"Object level precision: {object_precision:.4f}")
    print(f"Object level recall: {object_recall:.4f}")

    if out is not None:
        if not os.path.exists(out):
            with open(out, "w") as f:
                f.write(
                    "train_checkpoint,test_countries,pixel_level_iou,pixel_level_precision,pixel_level_recall,object_level_precision,object_level_recall\n"
                )
        with open(out, "a") as f:
            f.write(
                f"{model},{countries},{pixel_level_iou},{pixel_level_precision},{pixel_level_recall},{object_precision},{object_recall}\n"
            )



# Microsoft Planetary Computer API URL + collection id + bands for inference download command
MSPC_URL = "https://planetarycomputer.microsoft.com/api/stac/v1"
COLLECTION_ID = "sentinel-2-l2a"
BANDS_OF_INTEREST = ["B04", "B03", "B02", "B08"]

# Supported file formats for the inference polygon command
SUPPORTED_POLY_FORMATS_TXT = "Available file extensions: .parquet (GeoParquet, fiboa-compliant), .fgb (FlatGeoBuf), .gpkg (GeoPackage), .geojson / .json / .ndjson (GeoJSON)"

# List of all available countries
ALL_COUNTRIES = [
    "belgium",
    "cambodia",
    "croatia",
    "estonia",
    "portugal",
    "slovakia",
    "south_africa",
    "sweden",
    "austria",
    "brazil",
    "corsica",
    "denmark",
    "france",
    "india",
    "latvia",
    "luxembourg",
    "finland",
    "germany",
    "kenya",
    "lithuania",
    "netherlands",
    "rwanda",
    "slovenia",
    "spain",
    "vietnam",
]



from typing import Literal

import geopandas as gpd
import rasterio
import shapely.geometry
import shapely.ops
import torch
import torch.nn as nn
import torchvision.transforms.v2 as T

try:
    import ultralytics
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "ultralytics is not installed. Please install it with 'pip install ultralytics'."
    )


class DelineateAnything:
    """DelineateAnything model for delineating fields in satellite imagery."""

    checkpoints = {
        "DelineateAnything-S": "https://hf.co/torchgeo/delineate-anything-s/resolve/69cd440b0c5bd450ced145e68294aa9393ddae05/delineate_anything_s_rgb_yolo11n-b879d643.pt",
        "DelineateAnything": "https://hf.co/torchgeo/delineate-anything/resolve/60bea7b2f81568d16d5c75e4b5b06289e1d7efaf/delineate_anything_rgb_yolo11x-88ede029.pt",
    }
    transforms = nn.Sequential(
        T.Lambda(lambda x: x[:, :3, ...]),
        T.ToDtype(torch.float),
        T.Lambda(lambda x: x / 3000.0),
        T.Lambda(lambda x: x.clip(0, 1)),
        T.ConvertImageDtype(torch.float32),
    )

    def __init__(
        self,
        model: Literal[
            "DelineateAnything-S", "DelineateAnything"
        ] = "DelineateAnything-S",
        image_size: tuple[int, int] | int = 320,
        max_detections: int = 50,
        iou_threshold: float = 0.6,
        conf_threshold: float = 0.1,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
    ) -> None:
        """Initialize the DelineateAnything model.

        Args:
            model (str): The model variant to use, either "DelineateAnything-S" or "DelineateAnything".
            image_size (tuple[int, int] | int): The size of the input images. If an int is provided, it will be used for both width and height.
            max_detections (int): Maximum number of detections per image.
            iou_threshold (float): Intersection over Union threshold for filtering predictions.
            conf_threshold (float): Confidence threshold for filtering predictions.
            device (str): Device to run the model on, either "cuda" or "cpu".
        """
        super().__init__()
        self.image_size = (
            (image_size, image_size) if isinstance(image_size, int) else image_size
        )
        self.max_detections = max_detections
        self.iou_threshold = iou_threshold
        self.conf_threshold = conf_threshold
        self.device = device
        self.model = ultralytics.YOLO(self.checkpoints[model]).to(device)
        self.model.eval()
        self.transforms.eval()
        self.transforms = self.transforms.to(device)

    @staticmethod
    def polygonize(
        result: ultralytics.engine.results.Results,
        transform: rasterio.Affine,
        crs=rasterio.CRS,
    ) -> gpd.GeoDataFrame:
        """Convert the model predictions to a GeoDataFrame of georeferenced polygons.

        Args:
            result (ultralytics.engine.results.Results): The results from the model prediction.
            transform (rasterio.Affine): The affine transformation to convert pixel coordinates to geographic coordinates.
            crs (rasterio.CRS): The coordinate reference system of the output GeoDataFrame.

        Returns:
            gpd.GeoDataFrame: A GeoDataFrame containing the polygons of the delineated fields.
        """

        def pixel_to_geo(x, y, z=None):
            return transform * (x, y)

        df = result.to_df()
        df["geometry"] = df["segments"].apply(
            lambda x: shapely.geometry.Polygon(zip(x["y"], x["x"]))
        )
        df["geometry"] = df["geometry"].apply(
            lambda geom: shapely.ops.transform(pixel_to_geo, geom)
        )
        df.drop(["name", "class", "box", "segments"], axis=1, inplace=True)
        return gpd.GeoDataFrame(df, geometry=df["geometry"], crs=crs)

    def __call__(self, image: torch.Tensor) -> list[ultralytics.engine.results.Results]:
        """Forward pass through the model.

        Args:
            image (torch.Tensor): The input image tensor, expected to be in the format (B, C, H, W).

        Returns:
            list[ultralytics.engine.results.Results]: A list of results containing the model predictions.
        """
        if image.ndim == 3:
            image = image.unsqueeze(0)

        image = self.transforms(image.to(self.device))
        results = self.model.predict(
            image,
            imgsz=self.image_size,
            conf=self.conf_threshold,
            max_det=self.max_detections,
            iou=self.iou_threshold,
            device=self.device,
            half=True,
            verbose=False,
        )
        return results



"""FTW datamodule."""

from typing import Any, Optional

import kornia.augmentation as K
import torch
from matplotlib.figure import Figure
from torch.utils.data import Subset
from torchgeo.datamodules import NonGeoDataModule
from torchgeo.transforms import AugmentationSequential

from .datasets import FTW


def preprocess(sample):
    sample["image"] = sample["image"] / 3000
    return sample


class FTWDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the FTW dataset."""

    mean = torch.tensor([0, 0, 0, 0, 0, 0, 0, 0])
    std = torch.tensor([3000, 3000, 3000, 3000, 3000, 3000, 3000, 3000])

    def __init__(
        self,
        batch_size: int = 64,
        num_workers: int = 0,
        train_countries: list[str] = ["france"],
        val_countries: list[str] = ["france"],
        test_countries: list[str] = ["france"],
        temporal_options: str = "stacked",
        num_samples: int = -1,
        **kwargs: Any,
    ) -> None:
        """Initialize a new FTWDataModule instance.

        Note: you can pass train_batch_size, val_batch_size, test_batch_size to
            control the batch sizes of each DataLoader independently.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            train_countries: List of countries to use training splits from
            val_countries: List of countries to use validation splits from
            test_countries: List of countries to use test splits from
            **kwargs: Additional keyword arguments passed to
                :class:`~src.datasets.FTW`.
        """
        if "split" in kwargs:
            raise ValueError("Cannot specify split in FTWDataModule")

        self.train_countries = train_countries
        self.val_countries = val_countries
        self.test_countries = test_countries
        self.load_boundaries = kwargs.pop("load_boundaries", False)
        self.temporal_options = temporal_options
        self.num_samples = num_samples

        # for the temporal option windowA, windowB and median we will have 4 channel input
        if self.temporal_options in ("windowA", "windowB", "median"):
            self.mean = torch.tensor([0, 0, 0, 0])
            self.std = torch.tensor([3000, 3000, 3000, 3000])
        elif (
            self.temporal_options == "rgb"
        ):  # for the rgb temporal option we are just selecting these 3 channls from both window_a and window_b images
            self.mean = torch.tensor([0, 0, 0, 0, 0, 0])
            self.std = torch.tensor([3000, 3000, 3000, 3000, 3000, 3000])

        print("Loaded datamodule with:")
        print(f"Train countries: {self.train_countries}")
        print(f"Val countries: {self.val_countries}")
        print(f"Test countries: {self.test_countries}")
        print(f"Number of samples: {self.num_samples}")

        self.train_aug = AugmentationSequential(
            K.Normalize(mean=self.mean, std=self.std),
            K.RandomRotation(p=0.5, degrees=90),
            K.RandomHorizontalFlip(p=0.5),
            K.RandomVerticalFlip(p=0.5),
            K.RandomSharpness(p=0.5),
            data_keys=["image", "mask"],
        )
        self.aug = AugmentationSequential(
            K.Normalize(mean=self.mean, std=self.std), data_keys=["image", "mask"]
        )
        super().__init__(FTW, batch_size, num_workers, **kwargs)

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Called at the beginning of fit, validate, test, or predict. During distributed
        training, this method is called from every process across all the nodes. Setting
        state here is recommended.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        if stage in ["fit"]:
            self.train_dataset = FTW(
                countries=self.train_countries,
                split="train",
                load_boundaries=self.load_boundaries,
                temporal_options=self.temporal_options,
                num_samples=self.num_samples,
                **self.kwargs,
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = FTW(
                countries=self.val_countries,
                split="val",
                load_boundaries=self.load_boundaries,
                temporal_options=self.temporal_options,
                num_samples=self.num_samples,
                **self.kwargs,
            )
        if stage == "test":
            self.test_dataset = FTW(
                countries=self.test_countries,
                split="test",
                load_boundaries=self.load_boundaries,
                temporal_options=self.temporal_options,
                num_samples=self.num_samples,
                **self.kwargs,
            )

    # NOTE: can get rid of this when https://github.com/microsoft/torchgeo/pull/2003 is merged and released
    # TODO: remove this method, test, then pin torchgeo>=0.6
    def plot(self, *args: Any, **kwargs: Any) -> Optional[Figure]:
        """Run the plot method of the validation dataset if one exists.

        Should only be called during 'fit' or 'validate' stages as ``val_dataset``
        may not exist during other stages.

        Args:
            *args: Arguments passed to plot method.
            **kwargs: Keyword arguments passed to plot method.

        Returns:
            A matplotlib Figure with the image, ground truth, and predictions.
        """
        fig: Optional[Figure] = None
        dataset = self.train_dataset or self.val_dataset
        if isinstance(dataset, Subset):
            dataset = dataset.dataset
        if dataset is not None:
            if hasattr(dataset, "plot"):
                fig = dataset.plot(*args, **kwargs)
        return fig



import hashlib
import logging
import os

import numpy as np
import pandas as pd
import scipy
import scipy.stats
import xarray as xr

# Harvest day raster paths from https://github.com/ucg-uv/research_products/tree/main
SUMMER_START_RASTER_PATH = "assets/global_crop_calendar/sc_sos_3x3_v2_cog.tiff"
SUMMER_END_RASTER_PATH = "assets/global_crop_calendar/sc_eos_3x3_v2_cog.tiff"

logger = logging.getLogger()


def compute_md5(file_path):
    """Compute the MD5 checksum of a file."""
    hash_md5 = hashlib.md5()
    try:
        with open(file_path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
    except FileNotFoundError:
        return None
    return hash_md5.hexdigest()


def validate_checksums(checksum_file, root_directory):
    """Validate checksums stored in a checksum file."""
    if not os.path.isfile(checksum_file):
        print(f"Checksum file not found: {checksum_file}")
        return

    with open(checksum_file, "r") as f:
        lines = f.readlines()

    for line in lines:
        parts = line.strip().split()
        if len(parts) != 2:
            continue

        stored_checksum, file_path = parts
        file_path = os.path.join(root_directory, file_path)
        current_checksum = compute_md5(file_path)

        if current_checksum != stored_checksum:
            print("Checksum mismatch: {file_path}")
            return False
    return True


def harvest_to_datetime(harvest_day: int, year: int) -> pd.Timestamp:
    """
    Convert a harvest integer (day of the year) to a datetime object.

    Args:
        harvest_day (int): Day of the year (1-365).
        year (int): The year for which the date is to be calculated.

    Returns:
        pd.Timestamp: Corresponding datetime object.
    """
    return pd.to_datetime(f"{year}-{harvest_day}", format="%Y-%j")


# to-do func to get harvest integer from user provided bbox
def get_harvest_integer_from_bbox(
    bbox: list[int],
    start_year_raster_path: str = SUMMER_START_RASTER_PATH,
    end_year_raster_path: str = SUMMER_END_RASTER_PATH,
) -> list[int]:
    """
    Gets harvest integer from a user-provided bounding box. Note currently just uses summer crops.

    Args:
        bbox (str): Bounding box in the format 'minx,miny,maxx,maxy'.

    Returns:
        list: start and end harvest integer (day of the year).
    """

    start_harvest_dset = xr.open_dataset(start_year_raster_path, engine="rasterio")
    end_harvest_dset = xr.open_dataset(end_year_raster_path, engine="rasterio")

    # Clip the datasets to the bounding box
    start_value = start_harvest_dset.rio.clip_box(
        bbox[0], bbox[1], bbox[2], bbox[3], allow_one_dimensional_raster=True
    )

    if len(start_value["band_data"][0][0]) > 1:
        start_days = start_value["band_data"].values[0][0]
        logger.info(
            f"Multiple dates found in area of interest {start_days}. Using circular mean to determine harvest day."
        )
        start_value = int(round(scipy.stats.circmean(start_days, high=365, low=1)))
    else:
        start_value = int(start_value["band_data"].values[0][0][0])

    end_value = end_harvest_dset.rio.clip_box(
        bbox[0], bbox[1], bbox[2], bbox[3], allow_one_dimensional_raster=True
    )
    if len(end_value["band_data"][0][0]) > 1:
        end_days = end_value["band_data"].values[0][0]
        logger.info(
            f"Multiple dates found in area of interest {end_days}. Using circular mean to determine harvest day."
        )
        end_value = int(round(scipy.stats.circmean(end_days, high=365, low=1)))

    else:
        end_value = int(end_value["band_data"].values[0][0][0])

    return [start_value, end_value]



import os

from click.testing import CliRunner

from ftw_cli.cli import inference_download, inference_polygonize, inference_run


def test_inference_download():  # create_input
    runner = CliRunner()

    # Download imagery by ID from MS Planetary Computer
    inference_image = "inference_imagery/austria_example.tif"
    result = runner.invoke(
        inference_download,
        [
            "--win_a=S2B_MSIL2A_20210617T100559_R022_T33UUP_20210624T063729",
            "--win_b=S2B_MSIL2A_20210925T101019_R022_T33UUP_20210926T121923",
            "--bbox=13.0,48.0,13.2,48.2",
            "-o",
            inference_image,
            "-f",
        ],
    )
    assert result.exit_code == 0, result.output
    assert "Writing output" in result.output
    assert "Finished merging and writing output" in result.output
    assert os.path.exists(inference_image)


def test_inference_run():
    runner = CliRunner()

    # Check required files are present
    model_path = "3_Class_FULL_FTW_Pretrained.ckpt"
    assert os.path.exists(model_path)
    inf_input_path = "./src/tests/data-files/inference-img.tif"
    assert os.path.exists(inf_input_path)

    # Run inference
    inf_output_path = "austria_example_output_full.tif"
    result = runner.invoke(
        inference_run,
        [
            inf_input_path,
            "--model",
            model_path,
            "--out",
            inf_output_path,
            "--gpu",
            "0",
            "--overwrite",
            "--resize_factor",
            "2",
            "--overwrite",
        ],
    )
    assert result.exit_code == 0, result.output
    assert "Using custom trainer" in result.output
    assert "Finished inference and saved output" in result.output
    assert os.path.exists(inf_output_path)


def test_inference_polygonize():
    runner = CliRunner()

    # Check required files are present
    mask = "./src/tests/data-files/mask.tif"
    assert os.path.exists(mask)

    # Polygonize the file
    out_path = "polygons.gpkg"
    result = runner.invoke(inference_polygonize, [mask, "-o", out_path, "-f"])
    assert result.exit_code == 0, result.output
    assert "Polygonizing input file:" in result.output
    assert "Finished polygonizing output" in result.output
    assert os.path.exists(out_path)



import click
import wget

from .cfg import ALL_COUNTRIES, SUPPORTED_POLY_FORMATS_TXT
from .types import ModelVersions

# Imports are in the functions below to speed-up CLI startup time
# Some of the ML related imports (presumable torch) are very slow
# See https://github.com/fieldsoftheworld/ftw-baselines/issues/40


@click.group()
def ftw():
    """Fields of The World (FTW) - Command Line Interface"""
    pass


## Data group


@ftw.group()
def data():
    """Downloading, unpacking, and preparing the FTW dataset."""
    pass


@data.command("download", help="Download and unpack the FTW dataset.")
@click.option(
    "--out",
    "-o",
    type=str,
    default="./data",
    help="Folder where the files will be downloaded to. Defaults to './data'.",
)
@click.option(
    "--clean_download",
    "-f",
    is_flag=True,
    help="If set, the script will delete the root folder before downloading.",
)
@click.option(
    "--countries",
    type=str,
    default="all",
    help="Comma-separated list of countries to download. If 'all' (default) is passed, downloads all available countries. Available countries: "
    + ", ".join(ALL_COUNTRIES),
)
@click.option(
    "--no-unpack",
    is_flag=True,
    help="If set, the script will NOT unpack the downloaded files.",
)
def data_download(out, clean_download, countries, no_unpack):
    from ftw_cli.download_ftw import download
    from ftw_cli.unpack import unpack

    download(out, clean_download, countries)
    if not no_unpack:
        unpack(out)


@data.command(
    "unpack",
    help="Unpack the downloaded FTW dataset. Specify the folder where the data is located via INPUT. Defaults to './data'.",
)
@click.argument("input", type=str, default="./data")
def data_unpack(input):
    from ftw_cli.unpack import unpack

    unpack(input)


### Model group


@ftw.group()
def model():
    """Training and testing FTW models."""
    pass


@model.command("fit", help="Fit the model")
@click.option(
    "--config",
    "-c",
    required=True,
    type=click.Path(exists=True),
    help="Path to the config file (required)",
)
@click.option(
    "--ckpt_path",
    type=click.Path(exists=True),
    help="Path to a checkpoint file to resume training from",
)
@click.argument(
    "cli_args", nargs=-1, type=click.UNPROCESSED
)  # Capture all remaining arguments
def model_fit(config, ckpt_path, cli_args):
    from ftw_cli.model import fit

    fit(config, ckpt_path, cli_args)


@model.command("test", help="Test the model")
@click.option("--model", "-m", required=True, type=str, help="Path to model checkpoint")
@click.option("--dir", type=str, default="data/ftw", help="Directory of dataset")
@click.option("--gpu", type=int, default=0, help="GPU to use")
@click.option(
    "--countries",
    type=str,
    multiple=True,
    required=True,
    help="Countries to evaluate on",
)
@click.option(
    "--postprocess", is_flag=True, help="Apply postprocessing to the model output"
)
@click.option(
    "--iou_threshold",
    type=float,
    default=0.5,
    help="IoU threshold for matching predictions to ground truths",
)
@click.option(
    "--out", "-o", type=str, default="metrics.json", help="Output file for metrics"
)
@click.option(
    "--model_predicts_3_classes",
    is_flag=True,
    help="Whether the model predicts 3 classes or 2 classes",
)
@click.option(
    "--test_on_3_classes",
    is_flag=True,
    help="Whether to test on 3 classes or 2 classes",
)
@click.option(
    "--temporal_options",
    type=str,
    default="stacked",
    help="Temporal option (stacked, windowA, windowB, etc.)",
)
@click.argument(
    "cli_args", nargs=-1, type=click.UNPROCESSED
)  # Capture all remaining arguments
def model_test(
    model,
    dir,
    gpu,
    countries,
    postprocess,
    iou_threshold,
    out,
    model_predicts_3_classes,
    test_on_3_classes,
    temporal_options,
    cli_args,
):
    from ftw_cli.model import test

    test(
        model,
        dir,
        gpu,
        countries,
        postprocess,
        iou_threshold,
        out,
        model_predicts_3_classes,
        test_on_3_classes,
        temporal_options,
        cli_args,
    )


@model.command("download", help="Download model checkpoints")
@click.option(
    "--type",
    type=click.Choice(ModelVersions),
    required=True,
    help="Short model name corresponding to a .ckpt file in github.",
)
def model_download(type: ModelVersions):
    github_url = f"https://github.com/fieldsoftheworld/ftw-baselines/releases/download/v1/{type.value}"
    print(f"Downloading {github_url} to {type.value}")
    wget.download(github_url)


### Inference group

WIN_HELP = "URL to or Microsoft Planetary Computer ID of an Sentinel-2 L2A STAC item for the window {x} image"


@ftw.group()
def inference():
    """Inference-related commands."""
    pass


@inference.command(
    "download",
    help="Download 2 Sentinel-2 scenes & stack them in a single file for inference.",
)
@click.option("--win_a", type=str, required=True, help=WIN_HELP.format(x="A"))
@click.option("--win_b", type=str, required=True, help=WIN_HELP.format(x="B"))
@click.option(
    "--out", "-o", type=str, required=True, help="Filename to save results to"
)
@click.option(
    "--overwrite", "-f", is_flag=True, help="Overwrites the outputs if they exist"
)
@click.option(
    "--bbox",
    type=str,
    default=None,
    help="Bounding box to use for the download in the format 'minx,miny,maxx,maxy'",
)
def inference_download(win_a, win_b, out, overwrite, bbox):
    from ftw_cli.download_img import create_input

    create_input(win_a, win_b, out, overwrite, bbox)


@inference.command(
    "run",
    help="Run inference on the stacked Sentinel-2 L2A satellite images specified via INPUT.",
)
@click.argument("input", type=click.Path(exists=True), required=True)
@click.option(
    "--model",
    "-m",
    type=click.Path(exists=True),
    required=True,
    help="Path to the model checkpoint.",
)
@click.option(
    "--out",
    "-o",
    type=str,
    default=None,
    help="Output filename for the inference imagery. Defaults to the name of the input file name with 'inference.' prefix.",
)
@click.option(
    "--resize_factor",
    type=int,
    default=2,
    show_default=True,
    help="Resize factor to use for inference.",
)
@click.option(
    "--gpu",
    type=int,
    help="GPU ID to use. If not provided, CPU will be used by default.",
)
@click.option(
    "--patch_size",
    type=int,
    default=None,
    help="Size of patch to use for inference. Defaults to 1024 unless the image is < 1024x1024px.",
)
@click.option(
    "--batch_size", type=int, default=2, show_default=True, help="Batch size."
)
@click.option(
    "--padding",
    type=int,
    default=None,
    help="Pixels to discard from each side of the patch.",
)
@click.option(
    "--overwrite", "-f", is_flag=True, help="Overwrite outputs if they exist."
)
@click.option(
    "--mps_mode", is_flag=True, help="Run inference in MPS mode (Apple GPUs)."
)
def inference_run(
    input,
    model,
    out,
    resize_factor,
    gpu,
    patch_size,
    batch_size,
    padding,
    overwrite,
    mps_mode,
):
    from ftw_cli.inference import run

    run(
        input,
        model,
        out,
        resize_factor,
        gpu,
        patch_size,
        batch_size,
        padding,
        overwrite,
        mps_mode,
    )


@inference.command(
    "polygonize",
    help="Polygonize the output from inference for the raster image given via INPUT. Results are in the CRS of the given raster image.",
)
@click.argument("input", type=click.Path(exists=True), required=True)
@click.option(
    "--out",
    "-o",
    type=str,
    default=None,
    help="Output filename for the polygonized data. Defaults to the name of the input file with parquet extension. "
    + SUPPORTED_POLY_FORMATS_TXT,
)
@click.option(
    "--simplify",
    type=float,
    default=15,
    show_default=True,
    help="Simplification factor to use when polygonizing in the unit of the CRS, e.g. meters for Sentinel-2 imagery in UTM. Set to 0 to disable simplification.",
)
@click.option(
    "--min_size",
    type=float,
    default=500,
    show_default=True,
    help="Minimum area size in square meters to include in the output. Set to 0 to disable.",
)
@click.option(
    "--max_size",
    type=float,
    default=None,
    show_default=True,
    help="Maximum area size in square meters to include in the output. Disabled by default.",
)
@click.option(
    "--overwrite", "-f", is_flag=True, help="Overwrite outputs if they exist."
)
@click.option(
    "--close_interiors",
    is_flag=True,
    help="Remove the interiors holes in the polygons.",
)
def inference_polygonize(
    input, out, simplify, min_size, max_size, overwrite, close_interiors
):
    from ftw_cli.polygonize import polygonize

    polygonize(input, out, simplify, min_size, max_size, overwrite, close_interiors)


@inference.command(
    "filter_by_lulc", help="Filter the output raster in GeoTIFF format by LULC mask."
)
@click.argument("input", type=click.Path(exists=True), required=True)
@click.option(
    "--out",
    "-o",
    type=str,
    default=None,
    help="Output filename for the (filtered) polygonized data. Defaults to the name of the input file with parquet extension. "
    + SUPPORTED_POLY_FORMATS_TXT,
)
@click.option(
    "--overwrite", "-f", is_flag=True, help="Overwrite outputs if they exist."
)
@click.option(
    "--collection_name",
    type=str,
    default="io-lulc-annual-v02",
    help="Name of the LULC collection to use. Available collections: io-lulc-annual-v02 (default) and esa-worldcover",
)
@click.option(
    "--save_lulc_tif",
    is_flag=True,
    help="Save the LULC mask as a GeoTIFF.",
)
def inference_lulc_filtering(input, out, overwrite, collection_name, save_lulc_tif):
    from ftw_tools.postprocess.lulc_filtering import lulc_filtering

    lulc_filtering(
        input=input,
        out=out,
        overwrite=overwrite,
        collection_name=collection_name,
        save_lulc_tif=save_lulc_tif,
    )


if __name__ == "__main__":
    ftw()



"""Trainer for semantic segmentation."""

import warnings
from typing import Any, Optional, Union

import lightning
import matplotlib.pyplot as plt
import segmentation_models_pytorch as smp
import torch
import torch.nn as nn
import torchvision.models as models
from matplotlib.figure import Figure
from torch import Tensor
from torch.optim import AdamW
from torch.optim.lr_scheduler import CosineAnnealingLR
from torchgeo.datasets import unbind_samples
from torchgeo.models import FCN
from torchgeo.trainers.base import BaseTask
from torchmetrics import MetricCollection
from torchmetrics.classification import MulticlassAccuracy, MulticlassJaccardIndex
from torchvision.models._api import WeightsEnum


class CustomSemanticSegmentationTask(BaseTask):
    """Semantic Segmentation.

    This is currently a copy of torchgeo.trainers.SemanticSegmentationTask, but with a
    fix to allow loading of saved weights.
    """

    def __init__(
        self,
        model: str = "unet",
        backbone: str = "resnet50",
        weights: Optional[Union[WeightsEnum, str, bool]] = None,
        in_channels: int = 3,
        num_classes: int = 1000,
        num_filters: int = 3,
        loss: str = "ce",
        class_weights: Optional[list] = None,
        ignore_index: Optional[int] = None,
        lr: float = 1e-3,
        patience: int = 10,
        patch_weights: bool = False,
        freeze_backbone: bool = False,
        freeze_decoder: bool = False,
        model_kwargs: dict[Any, Any] = dict(),
    ) -> None:
        """Inititalize a new SemanticSegmentationTask instance.

        Args:
            model: Name of the
                `smp <https://smp.readthedocs.io/en/latest/models.html>`__ model to use.
            backbone: Name of the `timm
                <https://smp.readthedocs.io/en/latest/encoders_timm.html>`__ or `smp
                <https://smp.readthedocs.io/en/latest/encoders.html>`__ backbone to use.
                Note: if using a DPT model, the backbone must be a supported timm encoder
                from the list `here <https://smp.readthedocs.io/en/latest/encoders_timm.html>`__
                such as `tu-resnet50` or `tu-vit_base_patch16_224`.
            weights: Initial model weights. Either a weight enum, the string
                representation of a weight enum, True for ImageNet weights, False or
                None for random weights, or the path to a saved model state dict. FCN
                model does not support pretrained weights. Pretrained ViT weight enums
                are not supported yet.
            in_channels: Number of input channels to model.
            num_classes: Number of prediction classes.
            num_filters: Number of filters. Only applicable when model='fcn'.
            loss: Name of the loss function, currently supports
                'ce', 'jaccard' or 'focal' loss.
            class_weights: Optional rescaling weight given to each
                class and used with 'ce' loss.
            ignore_index: Optional integer class index to ignore in the loss and
                metrics.
            lr: Learning rate for optimizer.
            patience: Patience for learning rate scheduler.
            freeze_backbone: Freeze the backbone network to fine-tune the
                decoder and segmentation head.
            freeze_decoder: Freeze the decoder network to linear probe
                the segmentation head.
            model_kwargs: Additional keyword arguments to pass to the model

        Warns:
            UserWarning: When loss='jaccard' and ignore_index is specified.

        .. versionchanged:: 0.3
           *ignore_zeros* was renamed to *ignore_index*.

        .. versionchanged:: 0.4
           *segmentation_model*, *encoder_name*, and *encoder_weights*
           were renamed to *model*, *backbone*, and *weights*.

        .. versionadded: 0.5
            The *class_weights*, *freeze_backbone*, and *freeze_decoder* parameters.

        .. versionchanged:: 0.5
           The *weights* parameter now supports WeightEnums and checkpoint paths.
           *learning_rate* and *learning_rate_schedule_patience* were renamed to
           *lr* and *patience*.
        """
        print("Using custom trainer")
        if ignore_index is not None and loss == "jaccard":
            warnings.warn(
                "ignore_index has no effect on training when loss='jaccard'",
                UserWarning,
            )

        self.weights = weights
        super().__init__()
        print(self.hparams)

    def configure_losses(self) -> None:
        """Initialize the loss criterion.

        Raises:
            ValueError: If *loss* is invalid.
        """
        loss: str = self.hparams["loss"]
        ignore_index = self.hparams["ignore_index"]
        class_weights = None
        if self.hparams["class_weights"] is not None:
            class_weights = torch.tensor(self.hparams["class_weights"])
        if loss == "ce":
            if self.hparams["class_weights"] is not None:
                class_weights = torch.tensor(self.hparams["class_weights"])
            else:
                class_weights = None
            ignore_value = -1000 if ignore_index is None else ignore_index
            self.criterion = nn.CrossEntropyLoss(
                ignore_index=ignore_value, weight=class_weights
            )
        elif loss == "jaccard":
            self.criterion = smp.losses.JaccardLoss(
                mode="multiclass", classes=self.hparams["num_classes"]
            )
        elif loss == "focal":
            self.criterion = smp.losses.FocalLoss(
                "multiclass", ignore_index=ignore_index, normalized=True
            )
        else:
            raise ValueError(
                f"Loss type '{loss}' is not valid. "
                "Currently, supports 'ce', 'jaccard' or 'focal' loss."
            )

    def configure_metrics(self) -> None:
        """Initialize the performance metrics."""
        num_classes: int = self.hparams["num_classes"]
        ignore_index: Optional[int] = self.hparams["ignore_index"]
        metrics = MetricCollection(
            [
                MulticlassAccuracy(
                    num_classes=num_classes,
                    ignore_index=ignore_index,
                    multidim_average="global",
                    average="micro",
                ),
                MulticlassJaccardIndex(
                    num_classes=num_classes, ignore_index=ignore_index, average="micro"
                ),
            ]
        )
        self.train_metrics = metrics.clone(prefix="train_")
        self.val_metrics = metrics.clone(prefix="val_")
        self.test_metrics = metrics.clone(prefix="test_")

    def configure_models(self) -> None:
        """Initialize the model.

        Raises:
            ValueError: If *model* is invalid.
        """
        model: str = self.hparams["model"]
        backbone: str = self.hparams["backbone"]
        weights = self.weights
        in_channels: int = self.hparams["in_channels"]
        num_classes: int = self.hparams["num_classes"]
        num_filters: int = self.hparams["num_filters"]
        model_kwargs: dict[Any, Any] = self.hparams["model_kwargs"]
        patch_weights: bool = self.hparams["patch_weights"]

        if model == "unet":
            self.model = smp.Unet(
                encoder_name=backbone,
                encoder_weights="imagenet" if weights is True else None,
                in_channels=in_channels,
                classes=num_classes,
                **model_kwargs,
            )
        elif model == "deeplabv3+":
            self.model = smp.DeepLabV3Plus(
                encoder_name=backbone,
                encoder_weights="imagenet" if weights is True else None,
                in_channels=in_channels,
                classes=num_classes,
                **model_kwargs,
            )
        elif model == "fcn":
            self.model = FCN(
                in_channels=in_channels, classes=num_classes, num_filters=num_filters
            )
        elif model == "upernet":
            self.model = smp.UPerNet(
                encoder_name=backbone,
                encoder_weights="imagenet" if weights is True else None,
                in_channels=in_channels,
                classes=num_classes,
                **model_kwargs,
            )
        elif model == "segformer":
            self.model = smp.Segformer(
                encoder_name=backbone,
                encoder_weights="imagenet" if weights is True else None,
                in_channels=in_channels,
                classes=num_classes,
            )
        elif model == "dpt":
            self.model = smp.DPT(
                encoder_name=backbone,
                encoder_weights="imagenet" if weights is True else None,
                in_channels=in_channels,
                classes=num_classes,
                **model_kwargs,
            )
        else:
            raise ValueError(
                f"Model type '{model}' is not valid. "
                "Currently, only supports 'unet', 'deeplabv3+', and 'fcn', 'upernet', 'segformer', and 'dpt'."
            )

        # Freeze backbone
        if self.hparams["freeze_backbone"] and model in ["unet", "deeplabv3+"]:
            for param in self.model.encoder.parameters():
                param.requires_grad = False

        # Freeze decoder
        if self.hparams["freeze_decoder"] and model in ["unet", "deeplabv3+"]:
            for param in self.model.decoder.parameters():
                param.requires_grad = False

        if patch_weights:
            self.transfer_weights(self.model, backbone)

    def training_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> Tensor:
        """Compute the training loss and additional metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.

        Returns:
            The loss tensor.
        """
        x = batch["image"]
        y = batch["mask"]
        y_hat = self(x)
        loss: Tensor = self.criterion(y_hat, y)
        self.log("train_loss", loss)
        self.train_metrics(y_hat, y)
        self.log_dict(self.train_metrics)
        return loss

    def validation_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        """Compute the validation loss and additional metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """
        x = batch["image"]
        y = batch["mask"]
        y_hat = self(x)
        loss: Tensor = self.criterion(y_hat, y)
        self.log("val_loss", loss)
        self.val_metrics(y_hat, y)
        self.log_dict(self.val_metrics)

        if (
            batch_idx < 10
            and hasattr(self.trainer, "datamodule")
            and hasattr(self.trainer.datamodule, "plot")
            and self.logger
            and hasattr(self.logger, "experiment")
            and hasattr(self.logger.experiment, "add_figure")
        ):
            datamodule = self.trainer.datamodule
            batch["prediction"] = y_hat.argmax(dim=1)
            for key in ["image", "mask", "prediction"]:
                batch[key] = batch[key].cpu()
            sample = unbind_samples(batch)[0]

            fig: Optional[Figure] = None
            fig = datamodule.plot(sample)

            if fig:
                summary_writer = self.logger.experiment
                summary_writer.add_figure(
                    f"image/{batch_idx}", fig, global_step=self.global_step
                )
                plt.close()

            if fig:
                for logger in self.loggers:
                    summary_writer = logger.experiment
                    if hasattr(summary_writer, "add_figure"):
                        summary_writer.add_figure(
                            f"image/{batch_idx}", fig, global_step=self.global_step
                        )
                plt.close()

    def test_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> None:
        """Compute the test loss and additional metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """
        x = batch["image"]
        y = batch["mask"]
        y_hat = self(x)
        loss: Tensor = self.criterion(y_hat, y)
        self.log("test_loss", loss)
        self.test_metrics(y_hat, y)
        self.log_dict(self.test_metrics)

    def predict_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> Tensor:
        """Compute the predicted class probabilities.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.

        Returns:
            Output predicted probabilities.
        """
        x = batch["image"]
        y_hat: Tensor = self(x).softmax(dim=1)
        return y_hat

    def configure_optimizers(
        self,
    ) -> "lightning.pytorch.utilities.types.OptimizerLRSchedulerConfig":
        """Initialize the optimizer and learning rate scheduler.

        Returns:
            Optimizer and learning rate scheduler.
        """
        optimizer = AdamW(self.parameters(), lr=self.hparams["lr"], amsgrad=True)
        scheduler = CosineAnnealingLR(
            optimizer, T_max=self.hparams["patience"], eta_min=1e-6
        )
        return {
            "optimizer": optimizer,
            "lr_scheduler": {"scheduler": scheduler, "monitor": self.monitor},
        }

    def on_train_epoch_start(self) -> None:
        lr = self.optimizers().param_groups[0]["lr"]
        self.logger.experiment.add_scalar("lr", lr, self.current_epoch)

    def transfer_weights(self, model, backbone):
        base_model = None
        if backbone == "resnet18":
            base_model = models.resnet18(pretrained=True)
        elif backbone == "resnet50":
            base_model = models.resnet50(pretrained=True)
        elif backbone == "resnext50_32x4d":
            base_model = models.resnext50_32x4d(pretrained=True)

        if not base_model:
            print(
                "Pretrained weights for ",
                backbone,
                " not found. Unable to patch wieights",
            )
            return
        prefix = "encoder."
        pretrained_weights = base_model.state_dict()
        model_dict = model.state_dict()
        pretrained_dict = {}
        weights_ = 0
        update_weights = True

        for index, layer_key in enumerate(pretrained_weights):
            # TODO : generalizing the patch mapping
            encoder_key = prefix + layer_key
            layer_w = pretrained_weights[layer_key]
            if encoder_key in model_dict:
                if index == 0:  # pacth first conv. layer weights
                    # Extract pre-trained weights for the first convolutional layer
                    pretrained_conv1_weights = layer_w
                    # Retrieve the current conv1 weights
                    new_conv1_weights = model_dict[encoder_key]
                    new_conv1_weights[:, :3, :, :] = pretrained_conv1_weights[
                        :, :3, :, :
                    ]
                    new_conv1_weights[:, 4:7, :, :] = pretrained_conv1_weights[
                        :, :3, :, :
                    ]
                    print(
                        encoder_key,
                        " First layer: ",
                        model_dict[encoder_key].size(),
                        "=>",
                        new_conv1_weights.size(),
                    )
                    pretrained_dict[encoder_key] = new_conv1_weights
                else:
                    if model_dict[encoder_key].size() != layer_w.size():
                        print("Invalid size match for ", encoder_key)
                        update_weights = False
                        break
                    pretrained_dict[encoder_key] = layer_w
                weights_ += 1
        if update_weights:
            print("Updated weights_ count ", weights_)
            model_dict.update(pretrained_dict)
            model.load_state_dict(model_dict)
        else:
            print("Due to mismatch in the Tensor size, unable to patch weights.")



import multiprocessing
import os
import shutil
import sys
import zipfile

from tqdm import tqdm


def unpack(input):
    ftw_folder_path = os.path.join(input, "ftw")

    # Ensure the input folder exists
    if not os.path.exists(input):
        print(f"Folder {input} does not exist.")
        sys.exit(1)

    clean_and_create_ftw_folder(ftw_folder_path)
    unpack_zip_files(input, ftw_folder_path)


def clean_and_create_ftw_folder(ftw_folder_path):
    """
    Delete the existing ftw folder if it exists and recreate it.

    :param ftw_folder_path: Path to the ftw folder.
    """
    if os.path.exists(ftw_folder_path):
        print(f"Deleting existing folder {ftw_folder_path}...")
        shutil.rmtree(ftw_folder_path)

    os.makedirs(ftw_folder_path, exist_ok=True)
    print(f"Created new folder {ftw_folder_path}.")


def unpack_zip_file(zip_file_path, extract_folder_path):
    """
    Unpack a single .zip file into the specified folder.

    :param zip_file_path: Path to the .zip file.
    :param extract_folder_path: Folder where the contents will be extracted.
    """
    with zipfile.ZipFile(zip_file_path, "r") as zip_ref:
        zip_ref.extractall(extract_folder_path)
    return zip_file_path


def unpack_zip_files(root_folder_path, ftw_folder_path):
    """
    Unpack all .zip files in the given root folder into subfolders under the ftw folder using multiprocessing.

    :param root_folder_path: Path to the root folder where the .zip files are located.
    :param ftw_folder_path: Path to the ftw folder where unpacked files will be placed.
    """
    # Find all .zip files in the root folder
    zip_files = [f for f in os.listdir(root_folder_path) if f.endswith(".zip")]

    if not zip_files:
        print(f"No .zip files found in {root_folder_path}")
        return

    # Prepare for multiprocessing
    cpu_count = multiprocessing.cpu_count()
    print(f"Using {cpu_count} CPUs for unpacking.")

    # Define tasks for each zip file
    tasks = []
    for zip_file in zip_files:
        zip_file_path = os.path.join(root_folder_path, zip_file)
        folder_name = os.path.splitext(zip_file)[
            0
        ]  # Remove .zip extension to get folder name
        extract_folder_path = os.path.join(ftw_folder_path, folder_name)

        # Create the folder where contents will be extracted
        os.makedirs(extract_folder_path, exist_ok=True)

        tasks.append((zip_file_path, extract_folder_path))

    # Unpack the .zip files using multiprocessing with tqdm
    with multiprocessing.Pool(cpu_count) as pool:
        for _ in tqdm(
            pool.starmap(unpack_zip_file, tasks),
            total=len(tasks),
            desc="Unpacking files",
        ):
            pass



import pandas as pd
import pytest

from ftw.utils import get_harvest_integer_from_bbox, harvest_to_datetime


@pytest.fixture()
def bbox_iowa_aoi():
    return [-93.68708939, 41.9530844, -93.64078526, 41.98070608]


@pytest.fixture()
def bbox_kenya_aoi():
    return [34.71339499, -2.84654017, 40.24044535, 3.1539109]


def test_get_harvest_integer_from_bbox(bbox_iowa_aoi):
    # Test for a valid bounding box
    result = get_harvest_integer_from_bbox(bbox_iowa_aoi)
    assert isinstance(result, list)
    assert len(result) == 2  # Should return start and end harvest days
    assert result[0] == 88
    assert result[1] == 316


def test_get_harvest_integer_multi_harvest_days(bbox_kenya_aoi):
    # Test for a bounding box with multiple harvest days
    result = get_harvest_integer_from_bbox(bbox_kenya_aoi)
    assert isinstance(result, list)
    assert len(result) == 2  # Should return start and end harvest days
    assert result[0] == 243  # Example expected value, adjust as needed
    assert result[1] == 44  # Example expected value, adjust as needed


def test_harvest_to_datetime():
    # Test for a valid harvest day and year
    result = harvest_to_datetime(100, 2023)
    assert result == pd.Timestamp("2023-04-10")



import enum


class ModelVersions(enum.StrEnum):
    """Mapping from short_name to .ckpt file in github."""

    TWO_CLASS_CCBY = "2_Class_CCBY_FTW_Pretrained.ckpt"
    TWO_CLASS_FULL = "2_Class_FULL_FTW_Pretrained.ckpt"
    THREE_CLASS_CCBY = "3_Class_CCBY_FTW_Pretrained.ckpt"
    THREE_CLASS_FULL = "3_Class_FULL_FTW_Pretrained.ckpt"



import math
import os
import time

import kornia.augmentation as K
import numpy as np
import rasterio
import rasterio.features
import torch
from kornia.constants import Resample
from rasterio.enums import ColorInterp
from torch.utils.data import DataLoader
from torchgeo.datasets import stack_samples
from torchgeo.samplers import GridGeoSampler
from tqdm import tqdm

from ftw.datamodules import preprocess
from ftw.datasets import SingleRasterDataset
from ftw.trainers import CustomSemanticSegmentationTask


def run(
    input,
    model,
    out,
    resize_factor,
    gpu,
    patch_size,
    batch_size,
    padding,
    overwrite,
    mps_mode,
):
    if not out:
        out = os.path.join(
            os.path.dirname(input), "inference." + os.path.basename(input)
        )

    # IO related sanity checks
    assert os.path.exists(model), f"Model file {model} does not exist."
    assert model.endswith(".ckpt"), "Model file must be a .ckpt file."
    assert os.path.exists(input), f"Input file {input} does not exist."
    assert input.endswith(".tif") or input.endswith(".vrt"), (
        "Input file must be a .tif or .vrt file."
    )
    assert overwrite or not os.path.exists(out), (
        f"Output file {out} already exists. Use -f to overwrite."
    )

    # Determine the device: GPU, MPS, or CPU
    if mps_mode:
        assert torch.backends.mps.is_available(), "MPS mode is not available."
        device = torch.device("mps")
    elif gpu is not None and torch.cuda.is_available():
        device = torch.device(f"cuda:{gpu}")
    else:
        print("Neither GPU nor MPS mode is enabled, defaulting to CPU.")
        device = torch.device("cpu")

    # Load the input raster
    with rasterio.open(input) as src:
        input_height, input_width = src.shape
        print(f"Input image size: {input_height}x{input_width} pixels (HxW)")
        profile = src.profile
        transform = profile["transform"]
        tags = src.tags()

    # Determine the default patch size
    if patch_size is None:
        steps = [1024, 512, 256, 128]
        for step in steps:
            if step <= min(input_height, input_width):
                patch_size = step
                break
    print("Patch size:", patch_size)
    assert patch_size is not None, "Input image is too small"
    assert patch_size % 32 == 0, "Patch size must be a multiple of 32."
    assert patch_size <= min(input_height, input_width), (
        "Patch size must not be larger than the input image dimensions."
    )

    if padding is None:
        # 64 for patch sizes >= 1024, otherwise smaller paddings
        padding = math.ceil(min(1024, patch_size) / 16)
    print("Padding:", padding)

    stride = patch_size - padding * 2
    assert stride > 64, (
        "Patch size minus two times the padding must be greater than 64."
    )

    # Load task
    tic = time.time()
    task = CustomSemanticSegmentationTask.load_from_checkpoint(
        model, map_location="cpu"
    )
    task.freeze()
    model = task.model.eval().to(device)

    if mps_mode:
        up_sample = K.Resize(
            (patch_size * resize_factor, patch_size * resize_factor)
        ).to("cpu")
        down_sample = (
            K.Resize((patch_size, patch_size), resample=Resample.NEAREST.name)
            .to(device)
            .to("cpu")
        )
    else:
        up_sample = K.Resize(
            (patch_size * resize_factor, patch_size * resize_factor)
        ).to(device)
        down_sample = K.Resize(
            (patch_size, patch_size), resample=Resample.NEAREST.name
        ).to(device)

    dataset = SingleRasterDataset(input, transforms=preprocess)
    sampler = GridGeoSampler(dataset, size=patch_size, stride=stride)
    dataloader = DataLoader(
        dataset,
        sampler=sampler,
        batch_size=batch_size,
        num_workers=6,
        collate_fn=stack_samples,
    )

    # Run inference
    output_mask = np.zeros((input_height, input_width), dtype=np.uint8)
    dl_enumerator = tqdm(dataloader)

    for batch in dl_enumerator:
        images = batch["image"].to(device)
        images = up_sample(images)

        # torchgeo>=0.6 refers to the bounding box as "bounds" instead of "bbox"
        if "bounds" in batch and batch["bounds"] is not None:
            bboxes = batch["bounds"]
        else:
            bboxes = batch["bbox"]

        with torch.inference_mode():
            predictions = model(images)
            predictions = predictions.argmax(axis=1).unsqueeze(0)
            predictions = down_sample(predictions.float()).int().cpu().numpy()[0]

        for i in range(len(bboxes)):
            bb = bboxes[i]
            left, top = ~transform * (bb.minx, bb.maxy)
            right, bottom = ~transform * (bb.maxx, bb.miny)
            left, right, top, bottom = (
                int(np.round(left)),
                int(np.round(right)),
                int(np.round(top)),
                int(np.round(bottom)),
            )
            pleft = left + padding
            pright = right - padding
            ptop = top + padding
            pbottom = bottom - padding
            destination_height, destination_width = output_mask[
                ptop:pbottom, pleft:pright
            ].shape
            inp = predictions[i][
                padding : padding + destination_height,
                padding : padding + destination_width,
            ]
            output_mask[ptop:pbottom, pleft:pright] = inp

    # Save predictions
    profile.update(
        {
            "driver": "GTiff",
            "count": 1,
            "dtype": "uint8",
            "compress": "lzw",
            "nodata": 0,
            "blockxsize": 512,
            "blockysize": 512,
            "tiled": True,
            "interleave": "pixel",
        }
    )

    with rasterio.open(out, "w", **profile) as dst:
        dst.update_tags(**tags)
        dst.write_colormap(1, {1: (255, 0, 0), 2: (0, 255, 0)})
        dst.colorinterp = [ColorInterp.palette]
        dst.write(output_mask, 1)

    print(f"Finished inference and saved output to {out} in {time.time() - tic:.2f}s")



import math
import os
import re
import time

import fiona
import fiona.transform
import numpy as np
import rasterio
import rasterio.features
import shapely.geometry
from affine import Affine
from fiboa_cli.parquet import create_parquet, features_to_dataframe
from pyproj import CRS, Transformer
from shapely.ops import transform
from tqdm import tqdm

from .cfg import SUPPORTED_POLY_FORMATS_TXT


def polygonize(input, out, simplify, min_size, max_size, overwrite, close_interiors):
    """Polygonize the output from inference."""

    print(f"Polygonizing input file: {input}")

    # TODO: Get this warning working right, based on the CRS of the input file
    # if simplify is not None and simplify > 1:
    #    print("WARNING: You are passing a value of `simplify` greater than 1 for a geographic coordinate system. This is probably **not** what you want.")

    if not out:
        out = os.path.splitext(input)[0] + ".parquet"

    if os.path.exists(out) and not overwrite:
        print(f"Output file {out} already exists. Use -f to overwrite.")
        return
    elif os.path.exists(out) and overwrite:
        os.remove(out)  # GPKGs are sometimes weird about overwriting in-place

    tic = time.time()
    rows = []
    schema = {
        "geometry": "Polygon",
        "properties": {
            "id": "str",
            "area": "float",
            "perimeter": "float",
        },
    }
    i = 1
    # read the input file as a mask
    with rasterio.open(input) as src:
        original_crs = src.crs.to_string()
        is_meters = src.crs.linear_units in ["m", "metre", "meter"]
        equal_area_crs = CRS.from_epsg(
            6933
        )  # Define the equal-area projection using EPSG:6933
        tags = src.tags()

        input_height, input_width = src.shape
        mask = (src.read(1) == 1).astype(np.uint8)
        polygonization_stride = 2048
        total_iterations = math.ceil(input_height / polygonization_stride) * math.ceil(
            input_width / polygonization_stride
        )

        if out.endswith(".gpkg"):
            format = "GPKG"
        elif out.endswith(".parquet"):
            format = "Parquet"
        elif out.endswith(".fgb"):
            format = "FlatGeobuf"
        elif out.endswith(".geojson") or out.endswith(".json"):
            format = "GeoJSON"
        elif out.endswith(".ndjson"):
            format = "GeoJSONSeq"
        else:
            raise ValueError(
                "Output format not supported. " + SUPPORTED_POLY_FORMATS_TXT
            )

        is_geojson = format.startswith("GeoJSON")
        if is_geojson:
            epsg4326 = CRS.from_epsg(4326)
            affine = Transformer.from_crs(
                original_crs, epsg4326, always_xy=True
            ).transform

        rows = []
        with tqdm(total=total_iterations, desc="Processing mask windows") as pbar:
            for y in range(0, input_height, polygonization_stride):
                for x in range(0, input_width, polygonization_stride):
                    new_transform = src.transform * Affine.translation(x, y)
                    mask_window = mask[
                        y : y + polygonization_stride, x : x + polygonization_stride
                    ]
                    for geom_geojson, val in rasterio.features.shapes(
                        mask_window, transform=new_transform
                    ):
                        if val != 1:
                            continue

                        geom = shapely.geometry.shape(geom_geojson)

                        if close_interiors:
                            geom = shapely.geometry.Polygon(geom.exterior)
                        if simplify > 0:
                            geom = geom.simplify(simplify)

                        # Calculate the area of the reprojected geometry
                        if is_meters:
                            geom_proj_meters = geom
                        else:
                            # Reproject the geometry to the equal-area projection
                            # if the CRS is not in meters
                            geom_proj_meters = shapely.geometry.shape(
                                fiona.transform.transform_geom(
                                    original_crs, equal_area_crs, geom_geojson
                                )
                            )

                        area = geom_proj_meters.area
                        perimeter = geom_proj_meters.length

                        # Only include geometries that meet the minimum size requirement
                        if area < min_size or (
                            max_size is not None and area > max_size
                        ):
                            continue

                        if is_geojson:
                            geom = transform(affine, geom)

                        rows.append(
                            {
                                "geometry": shapely.geometry.mapping(geom),
                                "properties": {
                                    "id": str(i),
                                    "area": area * 0.0001,  # Add the area in hectares
                                    "perimeter": perimeter,  # Add the perimeter in meters
                                },
                            }
                        )
                        i += 1

                    pbar.update(1)

    if format == "Parquet":
        timestamp = tags.get("TIFFTAG_DATETIME", None)
        if timestamp is not None:
            pattern = re.compile(
                r"^(\d{4})[:-](\d{2})[:-](\d{2})[T\s](\d{2}):(\d{2}):(\d{2}).*$"
            )
            if pattern.match(timestamp):
                timestamp = re.sub(pattern, r"\1-\2-\3T\4:\5:\6Z", timestamp)
            else:
                print("WARNING: Unable to parse timestamp from TIFFTAG_DATETIME tag.")
                timestamp = None

        config = collection = {"fiboa_version": "0.2.0"}
        columns = ["geometry", "determination_method"] + list(
            schema["properties"].keys()
        )
        gdf = features_to_dataframe(rows, columns)
        gdf.set_crs(original_crs, inplace=True, allow_override=True)
        gdf["determination_method"] = "auto-imagery"
        if timestamp is not None:
            gdf["determination_datetime"] = timestamp
            columns.append("determination_datetime")

        create_parquet(gdf, columns, collection, out, config, compression="brotli")
    else:
        print(
            "WARNING: The fiboa-compliant GeoParquet output format is recommended for field boundaries."
        )
        if is_geojson:
            original_crs = epsg4326
        with fiona.open(out, "w", format, schema=schema, crs=original_crs) as dst:
            dst.writerecords(rows)

    print(f"Finished polygonizing output at {out} in {time.time() - tic:.2f}s")



import os

from click.testing import CliRunner

from ftw_cli.cli import data_download as download
from ftw_cli.cli import data_unpack as unpack


def test_data_download():
    runner = CliRunner()

    # Clean download
    result = runner.invoke(download, ["-f", "--countries=Rwanda", "--no-unpack"])
    assert result.exit_code == 0, result.output
    assert "Downloading selected countries: ['rwanda']" in result.output
    assert "Overall Download Progress: 100%" in result.output
    assert "Unpacking files:" not in result.output
    assert os.path.exists("data/rwanda.zip")
    assert not os.path.exists("data/ftw/rwanda")

    # Try again and expect to skip the download, but unpack files
    result = runner.invoke(download, ["--countries=Rwanda"])
    assert result.exit_code == 0, result.output
    assert "already exists, skipping download." in result.output
    assert "Unpacking files: 100%" in result.output
    assert os.path.exists("data/rwanda.zip")
    assert os.path.exists("data/ftw/rwanda")

    # Try again and expect to skip the download and not unpack
    result = runner.invoke(download, ["--countries=Rwanda"])
    assert result.exit_code == 0, result.output
    assert "already exists, skipping download." in result.output
    assert "Unpacking files:" in result.output


def test_data_unpack():
    runner = CliRunner()

    # Unpack the files
    result = runner.invoke(unpack, [])
    assert result.exit_code == 0, result.output
    assert "Unpacking files: 100%" in result.output

    # Error with non-existing folder
    result = runner.invoke(unpack, ["./invalid_folder"])
    assert result.exit_code == 1, result.output
    assert "Folder ./invalid_folder does not exist." in result.output



import rasterio.features
import shapely.geometry


def get_object_level_metrics(y_true, y_pred, iou_threshold=0.5):
    """Get object level metrics for a single mask / prediction pair.

    Args:
        y_true (np.ndarray): Ground truth mask.
        y_pred (np.ndarray): Predicted mask.
        iou_threshold (float, optional): IoU threshold for matching predictions to ground truths. Defaults to 0.5.

    Returns
        tuple (int, int, int): Number of true positives, false positives, and false negatives.
    """
    if iou_threshold < 0.5:
        raise ValueError(
            "iou_threshold must be greater than 0.5"
        )  # If we go lower than 0.5 then it is possible for a single prediction to match with multiple ground truths and we have to do de-duplication
    y_true_shapes = []
    for geom, val in rasterio.features.shapes(y_true):
        if val == 1:
            y_true_shapes.append(shapely.geometry.shape(geom))

    y_pred_shapes = []
    for geom, val in rasterio.features.shapes(y_pred):
        if val == 1:
            y_pred_shapes.append(shapely.geometry.shape(geom))

    tps = 0
    fns = 0
    tp_is = set()  # keep track of which of the true shapes are true positives
    tp_js = set()  # keep track of which of the predicted shapes are true positives
    fn_is = set()  # keep track of which of the true shapes are false negatives
    matched_js = set()
    for i, y_true_shape in enumerate(y_true_shapes):
        matching_j = None
        for j, y_pred_shape in enumerate(y_pred_shapes):
            if y_true_shape.intersects(y_pred_shape):
                intersection = y_true_shape.intersection(y_pred_shape)
                union = y_true_shape.union(y_pred_shape)
                iou = intersection.area / union.area
                if iou > iou_threshold:
                    matching_j = j
                    matched_js.add(j)
                    tp_js.add(j)
                    break
        if matching_j is not None:
            tp_is.add(i)
            tps += 1
        else:
            fn_is.add(i)
            fns += 1
    fps = len(y_pred_shapes) - len(matched_js)
    fp_js = (
        set(range(len(y_pred_shapes))) - matched_js
    )  # compute which of the predicted shapes are false positives

    # Create masks of the true positives, false positives, and false negatives
    # tp_i_mask = rasterio.features.rasterize([y_true_shapes[i] for i in tp_is], out_shape= y_true.shape)
    # tp_j_mask = rasterio.features.rasterize([y_pred_shapes[j] for j in tp_js], out_shape= y_pred.shape)
    # fp_j_mask = rasterio.features.rasterize([y_pred_shapes[j] for j in fp_js], out_shape= y_pred.shape)
    # fn_i_mask = rasterio.features.rasterize([y_true_shapes[i] for i in fn_is], out_shape= y_true.shape)

    return (tps, fps, fns)



from click.testing import CliRunner

from ftw_cli.cli import data_download as download
from ftw_cli.cli import data_unpack as unpack
from ftw_cli.cli import (
    inference_download,
    inference_polygonize,
    inference_run,
    model_fit,
    model_test,
)

# Run all the help functions for a basic CLI check
# Just ensures imports are correct and that there are no syntax errors in Python files etc


def test_data_download():
    runner = CliRunner()
    result = runner.invoke(download, ["--help"])
    assert result.exit_code == 0, result.output
    assert "Usage: download [OPTIONS]" in result.output


def test_data_unpack():
    runner = CliRunner()
    result = runner.invoke(unpack, ["--help"])
    assert result.exit_code == 0, result.output
    assert "Usage: unpack [OPTIONS] [INPUT]" in result.output


def test_inference_download():  # create_input
    runner = CliRunner()
    result = runner.invoke(inference_download, ["--help"])
    assert result.exit_code == 0, result.output
    assert "Usage: download [OPTIONS]" in result.output


def test_inference_polygonize():
    runner = CliRunner()
    result = runner.invoke(inference_polygonize, ["--help"])
    assert result.exit_code == 0, result.output
    assert "Usage: polygonize [OPTIONS] INPUT" in result.output


def test_inference_run():
    runner = CliRunner()
    result = runner.invoke(inference_run, ["--help"])
    assert result.exit_code == 0, result.output
    assert "Usage: run [OPTIONS] INPUT" in result.output


def test_model_fit():
    runner = CliRunner()
    result = runner.invoke(model_fit, ["--help"])
    assert result.exit_code == 0, result.output
    assert "Usage: fit [OPTIONS] [CLI_ARGS]..." in result.output


def test_model_test():
    runner = CliRunner()
    result = runner.invoke(model_test, ["--help"])
    assert result.exit_code == 0, result.output
    assert "Usage: test [OPTIONS] [CLI_ARGS]..." in result.output



import geopandas as gpd
import pytest
import rasterio
import torch

pytest.importorskip("ultralytics")


def test_delineate_anything():
    import ultralytics

    from ftw.models.delineate_anything import DelineateAnything

    device = "cpu"
    model = DelineateAnything(
        model="DelineateAnything-S",
        image_size=(320, 320),
        max_detections=50,
        iou_threshold=0.6,
        conf_threshold=0.1,
        device=device,
    )

    # test model inference
    x = torch.randn(2, 3, 256, 256, requires_grad=False, device=device)
    with torch.inference_mode():
        results = model(x)

    assert len(results) == 2
    assert isinstance(results[0], ultralytics.engine.results.Results)

    # test conversion of results to polygons
    transform = rasterio.Affine(1, 0, 0, 0, -1, 0)
    crs = rasterio.CRS.from_epsg(4326)
    gdf = model.polygonize(results[0], transform, crs)
    assert isinstance(gdf, gpd.GeoDataFrame)
    assert gdf.crs == crs



import hashlib
import logging
import os
import shutil
import time

import wget
from tqdm import tqdm

from .cfg import ALL_COUNTRIES


def load_checksums(local_md5_file_path):
    """
    Load the checksum data from a local md5 file.

    :param local_md5_file_path: Path to the local checksum.md5 file
    :return: Dictionary with country name as key and checksum hash as value
    """
    checksum_data = {}
    with open(local_md5_file_path, "r") as file:
        for line in file:
            country, checksum = line.strip().split(",")
            checksum_data[country.lower()] = checksum
    return checksum_data


def calculate_md5(file_path):
    """Calculate the MD5 checksum of a file."""
    hash_md5 = hashlib.md5()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def download(out, clean_download, countries):
    root_folder_path = os.path.abspath(out)

    # Deletes all files and directories in the root folder.
    if clean_download and os.path.exists(root_folder_path):
        try:
            shutil.rmtree(root_folder_path)
            print(f"Deleted old data from {root_folder_path}")
        except Exception as e:
            print(f"Failed to delete {root_folder_path}. Reason: {e}")

    # Ensure the root folder exists
    os.makedirs(root_folder_path, exist_ok=True)

    # Path for the log file inside the data folder
    log_file = os.path.join(root_folder_path, "download.log")

    # Initialize logging
    logging.basicConfig(
        filename=log_file,
        level=logging.INFO,
        format="%(asctime)s %(levelname)s: %(message)s",
    )
    logger = logging.getLogger()

    def custom_progress_bar(file_name):
        start_time = time.time()
        print(f"Downloading {file_name}...")

        def bar(current, total, width=None):
            if width is None:
                width = shutil.get_terminal_size((80, 20)).columns
            elapsed_time = time.time() - start_time
            speed = current / elapsed_time if elapsed_time > 0 else 0
            progress = current / total
            bar_width = int(
                (width - 40)
            )  # Adjust width for the text and make the bar 30% shorter
            block = int(round(bar_width * progress))
            progress_message = (
                f"{current / (1024**3):.3f} GB / {total / (1024**3):.3f} GB"
            )
            speed_message = f"Speed: {speed / (1024**2):.2f} MB/s"
            bar = f"[{'#' * block}{'-' * (bar_width - block)}]"
            print(f"{progress_message} {bar} {speed_message}", end="\r")

        return bar

    def download_file(url, local_file_path):
        """
        Downloads a file from the given URL using wget.

        :param url: URL of the file to be downloaded
        :param local_file_path: Path to the local file where the URL will be downloaded
        :return: True if download was successful, False otherwise
        """
        try:
            # Use wget to download the file with a custom progress bar
            wget.download(
                url,
                local_file_path,
                bar=custom_progress_bar(os.path.basename(local_file_path)),
            )
            logger.info(f"Downloaded {url} to {local_file_path}")
            print(f"\nDownloaded {local_file_path}")
            return True
        except Exception as e:
            logger.error(f"Error downloading {url}: {e}")
            print(f"Error downloading {url}: {str(e)}")
            return False

    def download_and_verify_md5(zip_file_path, country, checksum_data):
        """
        Download a .zip file for a country and verify its checksum.

        :param zip_file_path: Path to the local .zip file
        :param country: Name of the country
        :param checksum_data: Dictionary of country to checksum hash
        :return: True if checksum matches, False otherwise
        """
        print(f"Verifying {zip_file_path}...")

        if country not in checksum_data:
            print(f"No checksum found for {country}, skipping verification.")
            return False

        # Calculate the MD5 checksum of the downloaded file
        calculated_checksum = calculate_md5(zip_file_path)
        expected_checksum = checksum_data[country]

        if calculated_checksum == expected_checksum:
            print(f"Checksum verification passed for {country}.")
            logger.info(f"Checksum verification passed for {zip_file_path}")
            return True
        else:
            print(
                f"Checksum verification failed for {country}. Expected: {expected_checksum}, Found: {calculated_checksum}"
            )
            logger.error(f"Checksum verification failed for {zip_file_path}")
            return False

    def download_all_country_files(root_folder, country_names, checksum_data):
        """
        Download all .zip files for the specified countries and verify their checksums using basic URL downloads.

        :param root_folder: Path to the root folder where files will be downloaded
        :param country_names: List of country names whose files need to be downloaded
        :param checksum_data: Dictionary of country to checksum hash
        """
        base_url = "https://data.source.coop/kerner-lab/fields-of-the-world-archive/"

        # tqdm progress for the entire dataset
        overall_progress = tqdm(
            total=len(country_names), desc="Overall Download Progress", unit="country"
        )

        for country in country_names:
            url = f"{base_url}{country}.zip"
            local_file_path = os.path.join(root_folder, f"{country}.zip")

            # Check if the file already exists locally
            if os.path.exists(local_file_path):
                logger.info(
                    f"File {local_file_path} already exists, skipping download."
                )
                print(f"File {local_file_path} already exists, skipping download.")
            else:
                # Download the file
                if download_file(url, local_file_path):
                    # Verify the file checksum
                    download_and_verify_md5(
                        local_file_path, country.lower(), checksum_data
                    )

            # Update the overall progress
            overall_progress.update(1)

        overall_progress.close()

    # Main script
    # Step 1: Download the checksum.md5 file
    local_md5_file_path = os.path.join(root_folder_path, "checksum.md5")
    md5_file_url = (
        "https://data.source.coop/kerner-lab/fields-of-the-world-archive/checksum.md5"
    )

    if not os.path.exists(local_md5_file_path):
        if not download_file(md5_file_url, local_md5_file_path):
            print("Failed to download checksum.md5 file.")
            return
        else:
            print(f"Downloaded checksum.md5 to {local_md5_file_path}")

    # Step 2: Load the checksum data
    checksum_data = load_checksums(local_md5_file_path)

    # Step 3: Handle country selection (all or specific countries)
    if countries == "all":
        country_names = ALL_COUNTRIES
        print("Downloading all available countries...")
    else:
        country_names = [
            country.lower().strip()
            for country in countries.split(",")
            if country.lower().strip() in ALL_COUNTRIES
        ]
        print(f"Downloading selected countries: {country_names}")

    # Step 4: Download all .zip files for the specified countries and verify their checksums
    download_all_country_files(root_folder_path, country_names, checksum_data)



import os

import pytest
import torch
from click.testing import CliRunner

from ftw_cli.cli import data_download, model_fit, model_test

CKPT_FILE = "logs/FTW-CI/lightning_logs/version_0/checkpoints/last.ckpt"
CONFIG_FILE = "src/tests/data-files/min_config.yaml"


def test_model_fit():
    runner = CliRunner()

    # Download required data for the fit command
    runner.invoke(data_download, ["--countries=Rwanda"])
    assert os.path.exists("data/ftw/rwanda")
    assert os.path.exists(CONFIG_FILE)

    # Run minimal fit
    result = runner.invoke(model_fit, ["-c", CONFIG_FILE])
    assert result.exit_code == 0, result.output
    assert "Train countries: ['rwanda']" in result.output
    assert "`Trainer.fit` stopped: `max_epochs=1` reached." in result.output
    assert "Finished" in result.output
    assert os.path.exists(CKPT_FILE)


def test_model_test():
    runner = CliRunner()

    # Check model for Rwanda
    result = runner.invoke(
        model_test,
        [
            "--gpu",
            "0",
            "--model",
            CKPT_FILE,
            "--countries",
            "Rwanda",  # should be "rwanda", but let's test case insensitivity
            "--out",
            "results.csv",
        ],
    )
    assert result.exit_code == 0, result.output
    assert "Running test command" in result.output
    assert "Created dataloader" in result.output
    assert "Object level recall: 0.0000" in result.output
    assert os.path.exists("results.csv")


@pytest.mark.parametrize(
    "arch", ["unet", "deeplabv3+", "fcn", "segformer", "dpt", "upernet"]
)
@torch.inference_mode()
def test_model_archs(arch: str):
    from ftw.trainers import CustomSemanticSegmentationTask

    params = {
        "class_weights": [0.04, 0.08, 0.88],
        "loss": "ce",
        "backbone": "efficientnet-b3",
        "weights": False,
        "patch_weights": False,
        "in_channels": 8,
        "num_classes": 3,
        "num_filters": 64,
        "ignore_index": 3,
        "lr": 1e-3,
        "patience": 100,
        "model_kwargs": {},
    }
    params["model"] = arch

    if arch == "dpt":
        params["backbone"] = "tu-resnet18"

    model = CustomSemanticSegmentationTask(**params)
    model.eval()
    x = torch.randn(1, 8, 256, 256)
    y = model(x)
    assert y.shape == (1, 3, 256, 256), f"Output shape mismatch for {arch}: {y.shape}"



"""FTW dataset."""

import os
import random
from pathlib import Path
from typing import Any, Callable, Optional, Sequence, Union

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import rasterio
import torch
from matplotlib.figure import Figure
from torch import Tensor
from torchgeo.datasets import NonGeoDataset, RasterDataset

from .utils import validate_checksums


class SingleRasterDataset(RasterDataset):
    """A torchgeo dataset that loads a single raster file."""

    def __init__(self, fn: str, transforms: Optional[Callable] = None):
        """Initialize the SingleRasterDataset class.

        Args:
            fn (str): The path to the raster file.
            transforms (Optional[Callable], optional): The transforms to apply to the
                raster file. Defaults to None.
        """
        path = os.path.abspath(fn)
        self.filename_regex = os.path.basename(path)
        super().__init__(paths=os.path.dirname(path), transforms=transforms)


class FTW(NonGeoDataset):
    valid_countries = [
        "austria",
        "belgium",
        "brazil",
        "cambodia",
        "corsica",
        "croatia",
        "denmark",
        "estonia",
        "finland",
        "france",
        "germany",
        "india",
        "kenya",
        "latvia",
        "lithuania",
        "luxembourg",
        "netherlands",
        "portugal",
        "rwanda",
        "slovakia",
        "slovenia",
        "south_africa",
        "spain",
        "sweden",
        "vietnam",
    ]

    valid_splits = ["train", "val", "test"]

    def __init__(
        self,
        root: str = "data/ftw",
        countries: Union[Sequence[str], str] = None,
        split: str = "train",
        transforms: Optional[Callable[[dict[str, Any]], dict[str, Any]]] = None,
        checksum: bool = False,
        load_boundaries: bool = False,
        temporal_options: str = "stacked",
        num_samples: int = -1,
    ) -> None:
        """Initialize a new FTW dataset instance.

        Args:
            root: root directory where dataset can be found, this should contain the
                country folder
            countries: the countries to load the dataset from, e.g. "france"
            split: string specifying what split to load (e.g. "train", "val", "test")
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            checksum: if True, check the MD5 of the downloaded files (may be slow)
            load_boundaries: if True, load the 3 class masks with boundaries
            temporal_options : for abalation study, valid option are (stacked, windowA, windowB, median, rgb)
        Raises:
            AssertionError: if ``countries`` argument is invalid
            AssertionError: if ``split`` argument is invalid
            RuntimeError: if data is not found, or checksums don't match
        """
        self.root = root

        if countries is None:
            raise ValueError("Please specify the countries to load the dataset from")

        if isinstance(countries, str):
            countries = [countries]
        countries = [country.lower() for country in countries]
        for country in countries:
            assert country in self.valid_countries, f"Invalid country {country}"

        self.countries = countries
        assert split in self.valid_splits
        self.transforms = transforms
        self.checksum = checksum
        self.load_boundaries = load_boundaries
        self.temporal_options = temporal_options
        self.num_samples = num_samples

        if self.load_boundaries:
            print("Loading 3 Class Masks, with Boundaries")
        else:
            print("Loading 2 Class Masks, without Boundaries")

        print("Temporal option: ", temporal_options)

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found at root directory or corrupted.  Download dataset with `ftw data download`"
            )

        if checksum:
            assert self._checksum(), "Checksum of dataset does not match"

        self.filenames = []
        all_filenames = []

        for country in self.countries:
            country_root: str = os.path.join(self.root, country)
            chips_fn = os.path.join(country_root, f"chips_{country}.parquet")
            chips_df = gpd.read_parquet(str(chips_fn))
            chips_df = chips_df[chips_df["split"] == split]
            aoi_ids = chips_df["aoi_id"].values

            for idx in aoi_ids:
                window_b_fn = Path(
                    os.path.join(country_root, "s2_images/window_b", f"{idx}.tif")
                )
                window_a_fn = Path(
                    os.path.join(country_root, "s2_images/window_a", f"{idx}.tif")
                )
                masks_2c_fn = Path(
                    os.path.join(
                        country_root, "label_masks/semantic_2class", f"{idx}.tif"
                    )
                )
                masks_3c_fn = Path(
                    os.path.join(
                        country_root, "label_masks/semantic_3class", f"{idx}.tif"
                    )
                )

                # Skip the image AOI's which does not have all four corresponding files
                if not (
                    window_b_fn.exists()
                    and window_a_fn.exists()
                    and masks_2c_fn.exists()
                    and masks_3c_fn.exists()
                ):
                    continue

                if self.load_boundaries:
                    mask_fn = os.path.join(
                        country_root, "label_masks/semantic_3class", f"{idx}.tif"
                    )
                else:
                    mask_fn = os.path.join(
                        country_root, "label_masks/semantic_2class", f"{idx}.tif"
                    )

                if os.path.exists(mask_fn):
                    all_filenames.append(
                        {
                            "window_b": os.path.join(
                                country_root, "s2_images/window_b", f"{idx}.tif"
                            ),
                            "window_a": os.path.join(
                                country_root, "s2_images/window_a", f"{idx}.tif"
                            ),
                            "mask": mask_fn,
                        }
                    )

        if self.num_samples == -1:  # select all samples
            self.filenames = all_filenames
        else:
            self.filenames = random.sample(
                all_filenames, min(self.num_samples, len(all_filenames))
            )

        print("Selecting : ", len(self.filenames), " samples")

    def _checksum(self) -> bool:
        """Check the checksum of the dataset.

        Returns:
            True if the checksum matches, else False
        """
        for country in self.valid_countries:
            print(f"Validating checksums for {country}")
            for checksum_file in [
                "distances_checksums.md5",
                "masks_checksums.md5",
                "window_b_checksums.md5",
                "window_a_checksums.md5",
            ]:
                checksum_file = os.path.join(self.root, country, checksum_file)
                if not os.path.exists(checksum_file):
                    print(f"Checksum file {checksum_file} not found")
                    return False
                if not validate_checksums(checksum_file, self.root):
                    return False
        return True

    def _check_integrity(self) -> bool:
        """Check the integrity of the dataset structure.

        Returns:
            True if the dataset directories and split files are found, else False
        """

        for country in self.countries:
            if country not in self.valid_countries:
                print(f"Invalid country {country}")
                return False

            country_dir = os.path.join(self.root, country)
            if not os.path.exists(country_dir):
                print(f"Country directory {country_dir} not found")
                return False

            chips_fns = list(Path(country_dir).glob(f"chips_*.parquet"))
            # boundaries_fns = list(Path(country_dir).glob(f"boundaries_*.parquet"))
            if len(chips_fns) != 1:
                print(f"Country {country} does not have chips file")
                return False

            if self.load_boundaries:
                if not all(
                    [
                        os.path.exists(os.path.join(country_dir, "s2_images/window_b")),
                        os.path.exists(os.path.join(country_dir, "s2_images/window_a")),
                        os.path.exists(
                            os.path.join(country_dir, "label_masks/semantic_3class")
                        ),
                    ]
                ):
                    print(f"Country {country} does not have all required directories")
                    return False
            else:
                if not all(
                    [
                        os.path.exists(os.path.join(country_dir, "s2_images/window_b")),
                        os.path.exists(os.path.join(country_dir, "s2_images/window_a")),
                        os.path.exists(
                            os.path.join(country_dir, "label_masks/semantic_2class")
                        ),
                    ]
                ):
                    print(f"Country {country} does not have all required directories")
                    return False
        return True

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.filenames)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            dictionary containing "image" and "mask" PyTorch tensors
        """
        filenames = self.filenames[index]

        images = []

        if self.temporal_options in ("stacked", "median", "windowB", "rgb"):
            with rasterio.open(filenames["window_b"]) as f:
                window_b_img = f.read()
                if self.temporal_options == "rgb":  # select 3 channels only
                    window_b_img = window_b_img[:3]
                images.append(window_b_img)

        if self.temporal_options in ("stacked", "median", "windowA", "rgb"):
            with rasterio.open(filenames["window_a"]) as f:
                window_a_img = f.read()
                if self.temporal_options == "rgb":  # select 3 channels only
                    window_a_img = window_a_img[:3]
                images.append(window_a_img)

        if self.temporal_options == "median":
            images = np.array(images).astype(np.int32)
            image = np.median(images, axis=0).astype(np.int32)
        else:
            image = np.concatenate(images, axis=0).astype(np.int32)

        image = torch.from_numpy(image).float()

        with rasterio.open(filenames["mask"]) as f:
            mask = f.read(1)
        mask = torch.from_numpy(mask).long()

        sample = {"image": image, "mask": mask}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def plot(self, sample: dict[str, Tensor], suptitle: Optional[str] = None) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample return by :meth:`__getitem__`
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample
        """
        img1 = sample["image"][0:3].numpy().transpose(1, 2, 0)

        if self.temporal_options in (
            "stacked",
            "rgb",
        ):  # only two option where we will have more than 4 channels in input image
            img2 = (
                sample["image"][3:6]
                if self.temporal_options == "rgb"
                else sample["image"][4:7]
            )
            img2 = img2.numpy().transpose(1, 2, 0)

        mask = sample["mask"].numpy()
        num_panels = 3 if self.temporal_options in ("stacked", "rgb") else 2
        if "prediction" in sample:
            num_panels += 1
            predictions = sample["prediction"].numpy()

        fig, axs = plt.subplots(1, num_panels, figsize=(num_panels * 5, 8))
        axs = axs.flatten()
        axs[0].imshow(np.clip(img1, 0, 1))
        axs[0].axis("off")

        panel_id = 1
        if self.temporal_options in ("stacked", "rgb"):
            axs[panel_id].imshow(np.clip(img2, 0, 1))
            axs[panel_id].axis("off")
            axs[panel_id + 1].imshow(mask, vmin=0, vmax=2, cmap="gray")
            axs[panel_id + 1].axis("off")
            panel_id += 2
        else:
            axs[panel_id].imshow(mask, vmin=0, vmax=2, cmap="gray")
            axs[panel_id].axis("off")
            panel_id += 1

        if "prediction" in sample:
            axs[panel_id].imshow(predictions)
            axs[panel_id].axis("off")

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig
